@@ -20,14 +20,13 @@
  * TODO:
  * - Use termcap to fetch cursor up/down codes
  * - Better handling of write errors: also try sub blocks
- * - Support non-seekable in/output
  */
 
 #ifndef VERSION
 # define VERSION "(unknown)"
 #endif
 
-#define ID "$Id: dd_rescue.c,v 1.46 2004/08/28 21:45:09 garloff Exp $"
+#define ID "$Id: dd_rescue.c,v 1.52 2006/07/23 11:47:50 garloff Exp $"
 
 #ifndef SOFTBLOCKSIZE
 # define SOFTBLOCKSIZE 65536
@@ -56,11 +55,11 @@
 #include <sys/time.h>
 #include <sys/stat.h>
 
-int softbs, hardbs;
+int softbs, hardbs, syncfreq;
 int maxerr, nrerr, reverse, dotrunc, abwrerr, sparse, nosparse;
 int verbose, quiet, interact, force;
-char* buf;
-char *lname, *iname, *oname;
+void* buf;
+char *lname, *iname, *oname, *bbname = NULL;
 off_t ipos, opos, xfer, lxfer, sxfer, fxfer, maxxfer;
 
 int ides, odes, identical, pres;
@@ -76,14 +75,6 @@ const char* up = "\x1b[A"; //]
 const char* down = "\n";
 const char* right = "\x1b[C"; //]
 
-/* Large file support on kernel 2.4 glibc 2.1 systems */
-/* _FILE_OFFSET_BITS=64 should take care of it */
-#if 0 && defined(__GLIBC__) && __GLIBC__ == 2 && __GLIBC_MINOR__ == 1 && defined(O_LARGEFILE)
-const unsigned int olarge = O_LARGEFILE;
-#else
-const unsigned int olarge = 0;
-#endif
-
 inline float difftimetv(const struct timeval* const t2, 
 			const struct timeval* const t1)
 {
@@ -139,9 +130,9 @@ int openfile(const char* const fname, const int flags)
 	} else
 		fdes = open(fname, flags, 0640);
 	if (fdes == -1) {
-		char buf[128];
-		snprintf(buf, 128, "dd_rescue: (fatal): open \"%s\" failed", fname);
-		perror(buf); exit(17);
+		char nbuf[128];
+		snprintf(nbuf, 128, "dd_rescue: (fatal): open \"%s\" failed", fname);
+		perror(nbuf); exit(17);
 	}
 	return fdes;
 }
@@ -187,25 +178,29 @@ void doprint(FILE* const file, const int bs, const clock_t cl,
 			100.0*(cl-startclock)/(CLOCKS_PER_SEC*t1));
 }
 
-void printstatus(FILE* const file1, FILE* const file2, 
+void printstatus(FILE* const file1, FILE* const file2,
 		 const int bs, const int sync)
 {
 	float t1, t2; 
 	clock_t cl;
 
-	if (sync) 
-		fsync(odes);
+	if (file1 == stderr || file1 == stdout) 
+		fprintf(file1, "%s%s%s", up, up, up);
+	if (file2 == stderr || file2 == stdout) 
+		fprintf(file2, "%s%s%s", up, up, up);
+
+	if (sync) {
+		int err = fsync(odes);
+		if (err)
+			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!    \n", 
+			      oname, (float)ipos/1024, strerror(errno));
+	}
 
 	gettimeofday(&currenttime, NULL);
 	t1 = difftimetv(&currenttime, &starttime);
 	t2 = difftimetv(&currenttime, &lasttime);
 	cl = clock();
 
-	if (file1 == stderr || file1 == stdout) 
-		fprintf(file1, "%s%s%s", up, up, up);
-	if (file2 == stderr || file2 == stdout) 
-		fprintf(file2, "%s%s%s", up, up, up);
-
 	if (file1) 
 		doprint(file1, bs, cl, t1, t2, sync);
 	if (file2)
@@ -216,6 +211,17 @@ void printstatus(FILE* const file1, FILE* const file2,
 	}
 }
 
+void savebb( int block )
+{
+	fplog(stderr, "Bad block: %d\n", block);
+    
+	if( bbname == NULL )
+		return;
+	FILE *bbfile = fopen(bbname,"a");
+	fprintf(bbfile,"%d\n",block);
+	fclose(bbfile);
+}
+
 void printreport()
 {
 	/* report */
@@ -226,19 +232,38 @@ void printreport()
 	printstatus(report, logfd, 0, 1);
 }
 
-void cleanup()
+int cleanup()
 {
+	int rc, errs = 0;
 	if (odes != -1) {
 		/* Make sure, the output file is expanded to the last (first) position */
 		pwrite(odes, buf, 0, opos);
-		close(odes); 
+		rc = fsync(odes);
+		if (rc) {
+			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
+			      oname, (float)opos/1024, strerror(errno));
+			++errs;
+		}
+		rc = close(odes); 
+		if (rc) {
+			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
+			      oname, (float)opos/1024, strerror(errno));
+			++errs;
+		}
+	}
+	if (ides != -1) {
+		rc = close(ides);
+		if (rc) {
+			fplog(stderr, "dd_rescue: (warning): %s (%.1fk): %s!\n", 
+			      iname, (float)ipos/1024, strerror(errno));
+			++errs;
+		}
 	}
-	if (ides != -1)
-		close(ides);
 	if (logfd)
 		fclose(logfd);
 	if (buf)
 		free(buf);
+	return errs;
 }
 
 /* is the block zero ? */
@@ -383,6 +408,7 @@ int copyfile(const off_t max, const int bs)
 					 	*/
 					}
 				}
+				savebb(ipos/bs);
 				fxfer += toread; xfer += toread;
 				if (reverse) { 
 					ipos -= toread; opos -= toread; 
@@ -474,10 +500,11 @@ int copyfile(const off_t max, const int bs)
 				}
 			} /* rd > 0 */
 		} /* errno */
-		if (!quiet && !(xfer % (16*softbs)) && (xfer % (512*softbs))) 
+
+		if (syncfreq && !(xfer % (syncfreq*softbs)))
+			printstatus((quiet? 0: stderr), 0, bs, 1);
+		else if (!quiet && !(xfer % (16*softbs)))
 			printstatus(stderr, 0, bs, 0);
-		if (!quiet && !(xfer % (512*softbs))) 
-			printstatus(stderr, 0, bs, 1);
 	} /* remain */
 	return errs;
 }
@@ -540,27 +567,31 @@ void printversion()
 void printhelp()
 {
 	printversion();
-	fprintf(stderr, "dd_rescue copies data from one file (or block device) to another\n");
+	fprintf(stderr, "dd_rescue copies data from one file (or block device) to another.\n");
 	fprintf(stderr, "USAGE: dd_rescue [options] infile outfile\n");
 	fprintf(stderr, "Options: -s ipos    start position in  input file (default=0),\n");
-	fprintf(stderr, "         -S opos    start position in output file (def=ipos);\n");
+	fprintf(stderr, "         -S opos    start position in output file (def=ipos),\n");
 	fprintf(stderr, "         -b softbs  block size for copy operation (def=%i),\n", SOFTBLOCKSIZE );
-	fprintf(stderr, "         -B hardbs  fallback block size in case of errs (def=%i);\n", HARDBLOCKSIZE );
-	fprintf(stderr, "         -e maxerr  exit after maxerr errors (def=0=infinite);\n");
-	fprintf(stderr, "         -m maxxfer maximum amount of data to be transfered (def=0=inf);\n");
-	fprintf(stderr, "         -l logfdile name of a file to log errors and summary to (def=\"\");\n");
-	fprintf(stderr, "         -r         reverse direction copy (def=forward);\n");
-	fprintf(stderr, "         -t         truncate output file (def=no);\n");
-	fprintf(stderr, "         -d/D       use O_DIRECT for input/output (def=no);\n");
-	fprintf(stderr, "         -w         abort on Write errors (def=no);\n");
+	fprintf(stderr, "         -B hardbs  fallback block size in case of errs (def=%i),\n", HARDBLOCKSIZE );
+	fprintf(stderr, "         -e maxerr  exit after maxerr errors (def=0=infinite),\n");
+	fprintf(stderr, "         -m maxxfer maximum amount of data to be transfered (def=0=inf),\n");
+	fprintf(stderr, "         -y syncfrq frequency of fsync calls on outfile (def=512*softbs),\n");
+	fprintf(stderr, "         -l logfile name of a file to log errors and summary to (def=\"\"),\n");
+	fprintf(stderr, "         -o bbfile  name of a file to log bad blocks numbers (def=\"\"),\n");
+	fprintf(stderr, "         -r         reverse direction copy (def=forward),\n");
+	fprintf(stderr, "         -t         truncate output file (def=no),\n");
+#ifdef O_DIRECT
+	fprintf(stderr, "         -d/D       use O_DIRECT for input/output (def=no),\n");
+#endif
+	fprintf(stderr, "         -w         abort on Write errors (def=no),\n");
 	fprintf(stderr, "         -a         spArse file writing (def=no),\n");
-	fprintf(stderr, "         -A         Always write blocks, zeroed if err (def=no);\n");
-	fprintf(stderr, "         -i         interactive: ask before overwriting data (def=no);\n");
-	fprintf(stderr, "         -f         force: skip some sanity checks (def=no);\n");
-	fprintf(stderr, "         -p         preserve: preserve ownership / perms (def=no)\n");
+	fprintf(stderr, "         -A         Always write blocks, zeroed if err (def=no),\n");
+	fprintf(stderr, "         -i         interactive: ask before overwriting data (def=no),\n");
+	fprintf(stderr, "         -f         force: skip some sanity checks (def=no),\n");
+	fprintf(stderr, "         -p         preserve: preserve ownership / perms (def=no),\n");
 	fprintf(stderr, "         -q         quiet operation,\n");
-	fprintf(stderr, "         -v         verbose operation;\n");
-	fprintf(stderr, "         -V         display version and exit;\n");
+	fprintf(stderr, "         -v         verbose operation,\n");
+	fprintf(stderr, "         -V         display version and exit,\n");
 	fprintf(stderr, "         -h         display this help and exit.\n");
 	fprintf(stderr, "Note: Sizes may be given in units b(=512), k(=1024), M(=1024^2) or G(1024^3) bytes\n");
 	fprintf(stderr, "This program is useful to rescue data in case of I/O errors, because\n");
@@ -601,6 +632,8 @@ void breakhandler(int sig)
 int main(int argc, char* argv[])
 {
 	int c;
+	off_t syncsz = -1;
+	void **mp = (void **) &buf;
 
   	/* defaults */
 	softbs = SOFTBLOCKSIZE; hardbs = HARDBLOCKSIZE;
@@ -614,14 +647,16 @@ int main(int argc, char* argv[])
 	ides = -1; odes = -1; logfd = 0; nrerr = 0; buf = 0;
 	i_chr = 0; o_chr = 0;
 
-	while ((c = getopt(argc, argv, ":rtfihqvVwaAdDpb:B:m:e:s:S:l:")) != -1) {
+	while ((c = getopt(argc, argv, ":rtfihqvVwaAdDpb:B:m:e:s:S:l:o:y:")) != -1) {
 		switch (c) {
 			case 'r': reverse = 1; break;
 			case 't': dotrunc = O_TRUNC; break;
 			case 'i': interact = 1; force = 0; break;
 			case 'f': interact = 0; force = 1; break;
+#ifdef O_DIRECT
 			case 'd': o_dir_in  = O_DIRECT; break;
 			case 'D': o_dir_out = O_DIRECT; break;
+#endif
 			case 'p': pres = 1; break;
 			case 'a': sparse = 1; nosparse = 0; break;
 			case 'A': nosparse = 1; sparse = 0; break;
@@ -634,9 +669,11 @@ int main(int argc, char* argv[])
 			case 'B': hardbs = (int)readint(optarg); break;
 			case 'm': maxxfer = readint(optarg); break;
 			case 'e': maxerr = (int)readint(optarg); break;
+			case 'y': syncsz = readint(optarg); break;
 			case 's': ipos = readint(optarg); break;
 			case 'S': opos = readint(optarg); break;
 			case 'l': lname = optarg; break;
+			case 'o': bbname = optarg; break;
 			case ':': fplog (stderr, "dd_rescue: (fatal): option %c requires an argument!\n", optopt); 
 				printhelp();
 				exit(11); break;
@@ -669,6 +706,18 @@ int main(int argc, char* argv[])
 	}
 
 	/* sanity checks */
+#ifdef O_DIRECT
+	if ((o_dir_in || o_dir_out) && hardbs < sysconf(_SC_PAGESIZE)) {
+		hardbs = sysconf(_SC_PAGESIZE);
+		fplog(stderr, "dd_rescue: (warning): O_DIRECT requires hardbs of at least %i!\n",
+		      hardbs);
+	}
+
+	if (o_dir_in || o_dir_out)
+		fplog(stderr, "dd_rescue: (warning): We don't handle misalignment of last block w/ O_DIRECT!\n");
+				
+#endif				
+
 	if (softbs < hardbs) {
 		fplog(stderr, "dd_rescue: (warning): setting hardbs from %i to softbs %i!\n",
 		      hardbs, softbs);
@@ -676,19 +725,35 @@ int main(int argc, char* argv[])
 	}
 
 	if (hardbs <= 0) {
-		fplog(stderr, "dd_rescue: (fatal): you're crazy to set you block size to %i!\n", hardbs);
+		fplog(stderr, "dd_rescue: (fatal): you're crazy to set block size to %i!\n", hardbs);
 		cleanup(); exit(15);
 	}
 
+	/* Set sync frequency */
+	if (syncsz == -1)
+		syncfreq = 512;
+	else if (syncsz == 0)
+		syncfreq = 0;
+	else
+		syncfreq = (syncsz + softbs - 1) / softbs;
+
 	/* Have those been set by cmdline params? */
 	if (ipos == (off_t)-1) 
 		ipos = 0;
 
+#ifdef O_DIRECT
+	if (posix_memalign(mp, sysconf(_SC_PAGESIZE), softbs)) {
+		fplog(stderr, "dd_rescue: (fatal): allocation of aligned buffer failed!\n");
+		cleanup(); exit(18);
+	}
+#else
 	buf = malloc(softbs);
 	if (!buf) {
 		fplog(stderr, "dd_rescue: (fatal): allocation of buffer failed!\n");
 		cleanup(); exit(18);
 	}
+#endif
+
 	memset(buf, 0, softbs);
 
 	identical = check_identical(iname, oname);
@@ -697,7 +762,7 @@ int main(int argc, char* argv[])
 		cleanup(); exit(19);
 	}
 	/* Open input and output files */
-	ides = openfile(iname, O_RDONLY | olarge | o_dir_in);
+	ides = openfile(iname, O_RDONLY | o_dir_in);
 	if (ides < 0) {
 		fplog(stderr, "dd_rescue: (fatal): %s: %s\n", iname, strerror(errno));
 		cleanup(); exit(22);
@@ -706,7 +771,7 @@ int main(int argc, char* argv[])
 	/* Overwrite? */
 	/* Special case '-': stdout */
 	if (strcmp(oname, "-"))
-		odes = open(oname, O_WRONLY | olarge | o_dir_out, 0640);
+		odes = open(oname, O_WRONLY | o_dir_out, 0640);
 	else 
 		odes = 0;
 
@@ -726,7 +791,7 @@ int main(int argc, char* argv[])
 		}
 	}
 
-	odes = openfile(oname, O_WRONLY | O_CREAT | olarge | o_dir_out /*| O_EXCL*/ | dotrunc);
+	odes = openfile(oname, O_WRONLY | O_CREAT | o_dir_out /*| O_EXCL*/ | dotrunc);
 	if (odes < 0) {
 		fplog(stderr, "dd_rescue: (fatal): %s: %s\n", oname, strerror(errno));
 		cleanup(); exit(24);
